import java.util.*;

/** 
 * copyright Renqiang Min/Hojjat Ghaderi
 * KMerTrie class store the number of occurance of kmers in a sequence
 * (exact matach or with mismatch) compactly in a Trie structure.
 */

public class KMerTrie {
  
  private TrieNode root;   //root of the Trie
  
  private Alphabet alphabet; //Alphabet used in the sequence
  
  private int k;    //the lenght of kmers in this Trie
  
  private int mismatch;   //mismach number, 0 means exact match
  private String sequence; //the sequence  used to build the Trie
  
  /**
   * Create a KMerTrie for the given sequence based on
   * alphabet, kmer length, and mismatch number. <br>
   * Precontion: alphabet is nonempty and valid.
   * throws TrieException if any of the following conditions is violated:
   *    1. k>=1 
   *    2. sequence length must be >= k and uses only letters from the alphabet
   *    3. 0 <= m <= k, and 
   *     
   * @param alphabet The alphabet to store in the KMerTrie.
   * @param k The lenght of kmer
   * @param m the mismatch number, 0 means exact match
   * @param sequence The sequenece to be stored  
   */
  public KMerTrie(Alphabet alphabet, int k, int m, String sequence) 
    throws TrieException {
    this.root = new InternalNode(alphabet.size());
    this.alphabet = alphabet;
    this.k = k;
    this.mismatch = m;
    this.sequence = sequence;

   // TODO: complete this method. 
   //       First check to make sure all condition stated
   //       above hold, if not throw a proper exception.
   // Then you must build a Trie by adding all kmers in the sequence to
   // the root. You can use the "addKmer" method
   
  }
  
  /**
   * creating a string representation of trie structure
   * @return a prety print representation of the Trie in string form.
   */  
  public String toString() {
      return printKMerTrieHelper(root, "", 'r');
  }
  
  /**
   * PretyPrint the KMerTrie rooted at root, with indent preceding all output 
   * lines. The nodes are printed in pre-order.
   * @param r The root of the tree to be printed.
   * @param indent The string to go before output lines.
   * @param letter The letter that the root node corresponds to
   * @return prety print representation of the Trie in string form.
   */
  private String printKMerTrieHelper(TrieNode r, String indent, char letter) {
    String s = "";
    if (r == null) { //no child there
      return indent + "null\n";
    }

    // Pick a pretty indent.
    String newIndent;
    if (indent.equals("")) {
      newIndent = "...";
    }
    else { //every level we add new indentation
      newIndent = "..." + indent;
    }
    
    if (r instanceof LeafNode) { //we are at a leaf
      LeafNode lf = (LeafNode) r;
      s = s + indent + letter + ":" + Integer.toString(lf.getCount()) + "\n";
    }
    else { //we are at an internal node
      s = s + indent + letter + "\n";
      InternalNode internal = (InternalNode) r;
      //prety print the children 
      for (int i = 0; i < alphabet.size(); i ++) {
        s = s + printKMerTrieHelper(internal.getChild(i), newIndent, 
			alphabet.getLetter(i));
        //s = s + "\n";
      }
    }
    return s;
  }
  
  
  
  /**
   * adds kmer to the trie. If the kmer is already in the trie
   * increament its count.
   * Precondition: kmer has size k, root is not null
   * @param kmer The kmer to be added to the trie.
   */   
  private void addKmer(String kmer) {    
    if (mismatch == 0) { //using exect match
      addKmerWithoutMismatch(root, kmer, 0);
    } else { //allowing mismatch 
      addKmerWithMismatch(root, kmer, mismatch, 0);
    }   
  } 
  
  
  /**
   * Travese the trie starting in node r using the 
   * kmer (from index idx to the end) and add the kmer to the trie.
   * Preconditions: r is not null and corresponds to the node in the
   * trie as a result of following the kmer from index 0 to idx-1
   *   from root. Also, kmer has size k, idx is 0 to k-1.
   * @param r Starting node
   * @param kmer The kmer 
   * @param idx The index in the kmer
   */ 
  private void addKmerWithoutMismatch(TrieNode r, String kmer, int idx) {   
  //TODO: complete this method using recursion. 
  // Hint: if idx=k-1 the corresponding child will be a leaf. 
  // If the child is not null (i.e. kmer already in the Trie) we just 
  // update the count otherwise just make a new leaf with the kmer.
  // if idx<k-1 we need to follow the child corrsponding to
  // kmr[idx] in r (if the child is null we need  to make an iternal node 
  // for it first)) and call recursively.   
  
  }

  
  /**
   * Travese the trie starting in node r using the 
   * kmer (from index idx to the end) and add the kmer to the trie.
   * Also find any other kmer with the same prefix 0..idx-1 
   * that that has a mismatach less than m, and add those to the trie as well. 
   * 
   * Preconditions: r is not null and corresponds to the node in the
   * trie as a result of following the kmer from index 0 to idx-1
   *   from root. Also, kmer has size k, idx is 0 to k-1, 0<=m<=k.
   * @param r Starting node
   * @param kmer The kmer 
   * @param m The mismatch allowed for the rest of kmer.
   * @param idx The index in the kmer
   */
  private void addKmerWithMismatch(TrieNode r, String kmer, int m, int idx) {
  //TODO: complete this method using recursion. 
  // Hints: if m=0, means no mismatch allow, so can reuse ...
  // so let's assume m>0:
  // if idx=k-1 we are at the last level ready to add the leaf corresponding
  // to idx child. Moreovere we must add any othre kmers which has
  // any other letters from the alphabet in idx index (why?)
  // If such child is not null (i.e. such kmer already in the Trie) we just 
  // update the count otherwise just make a new leaf with the kmer.
  // Now if idx<k-1 we need to follow the child corrsponding to
  // kmr[idx] in r (if the child is null we need  to make an iternal node 
  // for it first)) and call recursively. but since m>1 we need to cosider
  // any other letter from alphabet in place or kmr[idx] and follow that
  // child recursivly (what will m be in the recursive call?) 
    
  }


  /**
   * compute the dot product between the trie and a given sequence 
   * (stored in a linkedlist). Note that the linked list is a compact
   * representation of feature vector as it only holds kmers with 
   * nonzero count.
   * 
   * Precondition: kmrList is not null
   * throws TrieException if kmrList has a kmer with a size different than
   * k in the trie or has a letter not in the alphabet.
   * @param kmerList A linkedList of kmers with their counts (this is
   * in effect a compact representaiton of a feature vector).
   * @return the dot product of the two sequences
   */ 
  public int dotProduct(LinkedList<LeafNode> kmerList) throws TrieException {
    //TODO: complte this method see the handout.
    //  Hint: use kmerCount method
    //you also need to check the conditions stated in the @throws and throw
    //a proper exception if necessary.
    
  }
  

  /**
   * traverse the trie from root to find the kmer and return 
   * its count, if it's not in the trie return 0.
   * precondition: kmer uses the same alphabet of trie and has size k.
   * @param kmer the kmer to find its count
   * @return the count value for the given kmer if it is in the trie, 
   * otherwise 0.
   */ 
  public int getKmerCount(String kmer) {
    //TODO: implement this method, 
    //  Hint: use a helper that has 
    //        root,kmer, and possiblly an index as  arguments....
    
  }
  
  

  /**
   * traverse the trie from root and create a list of all 
   * kmers in the trie.
   * @return a LinkedList of kmers in the trie. Each node
   * contians the Kmer and its count.
   */ 
  public LinkedList<LeafNode> getListofKmers() {
   //TODO: implement this method.
   //       Hint: use a helper that has as argument a root 
   //             and a LinkList that leaves are added to it
   //             as you travers the trie...
    
  }

 

}
